#include <stdio.h>
#include <stdlib.h>
#include <string.h>
 
#include "gzip.h"
#include "enc.h"
 
#define TRUE 1
#define FALSE 0

#define affirm(ex) ((ex) ? 0 : affirm_err((#ex), __FILE__, __LINE__))
int affirm_err(char *err, char *file, int line);

#define CKCK {fprintf(CK_fp, "Got to %d %s\n", __LINE__, __FILE__); ck_np++<9000?0:exit(0); }  /**/
int ck_np = 0;
extern FILE *CK_fp;  /**/
 
#define mEncs 10
#define nanoCryptStr "nanoCrypt"
enc_t encs[mEncs];
int nEncs = 0;
 
char enc_errMsg[256] = "";
int enc_errStatus = 0;  // 0 means no error.  Positive number means error.


void enc_register(enc_t enc)
{	int i;
	int encNum;
 
// Perform initialisation if not already done.
	if (nEncs == 0)
	{	for (i=0; i<mEncs; i++)
			encs[i].encNum = -1;
	}
 
// Check encNum in range.
	encNum = enc.encNum;
	if (encNum<0 || encNum>mEncs)  // 0 means no encryption.
	{	fprintf(stderr, "%s\n", "encBuf: encNum out of range!");
		exit(1);
	}
 
// Register the encryption.
	encs[encNum] = enc;
	nEncs++;
}


void enc_glue(void)
{	extern enc_t enc_dummy_details(void); enc_register(enc_dummy_details());
	extern enc_t enc_aes128_details(void); enc_register(enc_aes128_details());
	extern enc_t enc_aes256_details(void); enc_register(enc_aes256_details());
	extern enc_t enc_aes3_details(void); enc_register(enc_aes3_details());
}

// -----------------------
// Encryption Input Buffer
// -----------------------

struct eib_s 
{	int isVerbatim;
	unsigned short nBufsIn;
	unsigned char *buf;
	unsigned char *qEnd, *qPos;
	FILE *fin;
	int whichEnc;
	void *encObj;
	enc_t *enc;
	gzipPull_t *gz;
};


void eib_close(eib_t *eib)
{	if (eib->encObj != NULL)
	{	affirm(eib->enc != NULL);
		eib->enc->destroy(eib->encObj);
	}
	if (eib->buf != NULL)
		free(eib->buf);
	if (eib->gz)
		gzipPull_free(eib->gz);
	free(eib);
	enc_errStatus = 0;
}


int eib_keyInit(eib_t *eib, char *key)
{	affirm(eib->enc != NULL);
	eib->encObj = eib->enc->create(key);
	if (eib->encObj == NULL)
	{	enc_errStatus = 0;
		return FALSE;
	}
	else
	{	enc_errStatus = 1;
		return TRUE;
	}
}


int eib_whichEnc(eib_t *eib)
{	return eib->whichEnc;
}


inline int eib_getc(eib_t *eib)
{	if (eib->isVerbatim)
	{ // No encryption.
		return getc(eib->fin);
	}
	else if (eib->whichEnc == -1)
	{ // No encryption, but the initial part of the file is in the buffer.
	  // We must consume that before setting verbatim.
		if (eib->qPos < eib->qEnd)
			return *(eib->qPos++);	
		else
		{	eib->isVerbatim = TRUE;	
			return getc(eib->fin);
		}
	}
	else
	{ // Decrypt and decompress.
		int ch = gzipPull_getc(eib->gz);
		return ch;
	}
}


static int eib_getChar(void *inStream)
{	eib_t *eib = (eib_t*)inStream;
 
	if (eib->qPos < eib->qEnd)
	{ // Grab a queued character.
		enc_errStatus = 0;
		return *(eib->qPos++);	
	}
	{ // Empty queue, so then:- Read. Decrypt. Que.

		int nRead, inBufSiz, nBytes, bufSiz;
		unsigned char *buf;

	// Get the buffer.
		buf = eib->buf;
		bufSiz = eib->enc->bufSize;
		nRead = fread(buf, bufSiz, 1, eib->fin); 
		if (nRead == 0)
		{	enc_errStatus = 0;
			return EOF;
		}

	// Decrypt it.
		eib->enc->decBuf(eib->encObj, buf);

	// Undo Making the buffer different from a previous one with the same data.
		*(unsigned short*)buf = *(unsigned short*)buf ^ eib->nBufsIn++;

	// How many bytes are here really?
		if (bufSiz <= 256) 
		{	inBufSiz = bufSiz -1;
			nBytes = buf[inBufSiz];
		}
		else
		{	inBufSiz = bufSiz - 2;
			nBytes = *(unsigned short*)(&buf[inBufSiz]);
		}
		if (nBytes < 1)
		{	strcpy(enc_errMsg, "Corrupted input file 2");
			enc_errStatus = 1;
			return EOF;
		}
		if (nBytes > inBufSiz)
		{	strcpy(enc_errMsg, "Corrupted input file 3");
			enc_errStatus = 1;
			return EOF;
		}

	// Set up the queue.
		eib->qPos = buf;
		eib->qEnd = buf + nBytes;

	// Grab a queued character.
		enc_errStatus = 0;
		return *(eib->qPos++);	

	} // Empty Queue.
}


eib_t *eib_open(FILE *fin, int isPosEnc)
{	eib_t *eib;
	int cryptStrLen = strlen(nanoCryptStr) + 1;
	int ch;
	int nRead;
	unsigned char *buf;
 
	eib = malloc(sizeof(eib_t));
	if (eib == NULL)
	{	strcpy(enc_errMsg, "Memory allocation failure 1");
		enc_errStatus = 1;
		return NULL;
	}

	eib->fin = fin;
 
// Set resources to NULL.
	eib->gz = NULL;
	eib->buf = NULL;
	eib->enc = NULL;
	eib->encObj = NULL;
 
	if (!isPosEnc)
	{
	// No encryption.
		eib->whichEnc = -1;
		eib->isVerbatim = TRUE;
	}
	else
	{	
	// Allocate buffer just big enough to hold intro string.
		buf = malloc(cryptStrLen);
		if (buf == NULL)
		{	strcpy(enc_errMsg, "Memory allocation failure 2");
			enc_errStatus = 1;
			free(eib);
			return NULL;
		}
		eib->buf = buf;
	 
		nRead = fread(buf, 1, cryptStrLen, fin);
		if (nRead!=cryptStrLen || strcmp(buf,nanoCryptStr))
		{ // Its not an encrypted nano file.
			eib->whichEnc = -1;
 
		// Can unread the string, so queue it.
		// Set up queue, so as not to lose the characters read in.
			eib->qPos = buf;
			eib->qEnd = buf + nRead;

		// This is temporary, until this small buffer is consumed.
			eib->isVerbatim = FALSE;
		}
		else
		{	// This is supposedly an encrypted nano file.
			eib->whichEnc = -2;
			nRead = fscanf(fin, "%d", &eib->whichEnc);
			ch = fgetc(fin);
			if (ch != '\0')
			{ // Its not an encrypted nano file.
				eib->whichEnc = -1;
	 
			// Can unread the string, so queue it.
			// Set up queue, so as not to lose the characters read in.
				eib->qPos = buf;
				eib->qEnd = buf + nRead;

			// This is temporary, until this small buffer is consumed.
				eib->isVerbatim = FALSE;
			}
			else if (eib->whichEnc<0 || eib->whichEnc>=mEncs)
			{	sprintf(enc_errMsg, "Invalid encryption 1 %d", eib->whichEnc);
				enc_errStatus = 1;
				free(buf);
				free(eib);
				return NULL;
			}
			else
			{ // Looks like valid nanoEnc intro.
				eib->enc = & encs[eib->whichEnc];
				if (eib->enc->encNum != eib->whichEnc)
				{	sprintf(enc_errMsg, "Invalid encryption 2 %d", eib->whichEnc);
					enc_errStatus = 1;
					free(buf);
					free(eib);
					return NULL;
				}
				else
				{ // We really do seem to have an encrypted nano file and we know
				  // which encryption to use.
 
				// Reallocate the buffer.
					buf = realloc(buf, eib->enc->bufSize);
					if (buf == NULL)
					{	strcpy(enc_errMsg, "Memory allocation failure 3");
						enc_errStatus = 1;
						free(eib);
						return NULL;
					}
					eib->buf = buf;
 
				// Set up the queue, at the end with no bytes.
					eib->qPos = eib->qEnd = buf;
					eib->isVerbatim = FALSE;
					eib->nBufsIn = 0;

				// Set up the decompression.
					eib->gz = gzipPull_new(eib_getChar, (void*)eib, 1);
 
				// We are not yet ready to allocate the encryption because we dont have a key.
					eib->encObj = NULL;
				}
			}
		} // done reading file intro.
	} // isPosEnc.
 
	enc_errStatus = 0;
	return eib;
}


// ------------------------
// Encryption Output Buffer
// ------------------------

struct eob_s 
{	int isVerbatim;
	int bufSiz;
	int inBufSiz;
	unsigned short nBufsOut;
	unsigned char *buf;
	FILE *fout;
	int whichEnc;
	void *encObj;
	void (*encBuf)(void *encObj, unsigned char *buf);
	enc_t *enc;	
	unsigned char *bufs[2];
	long bufPos;
	int bufNum;
	gzipPush_t *gz;
};


static void completeEncWriteBlk
(	FILE *fout, 
	void (*encBuf)(void *encObj, unsigned char *buf),
	void *encObj,
	unsigned char *buf,
	unsigned short *nBufsOut,
	int bufSiz,
	int nBytes
)
{	int inBufSiz;
 
// Complete the buffer.
	if (bufSiz <= 256) 
	{	inBufSiz = bufSiz -1;
		buf[inBufSiz] = (unsigned char)nBytes;
	}
	else
	{	inBufSiz = bufSiz - 2;
		*(unsigned short*)(&buf[inBufSiz]) = (unsigned short)nBytes;
	}
	
// Zero out any undefined part of the buffer.
	memset(buf+nBytes, 0, inBufSiz-nBytes);
 
// Make the buffer different from a previous one with the same data.
	*(unsigned short*)buf = *(unsigned short*)buf ^ ((*nBufsOut)++);
 
// Encrypt the buffer.
	(encBuf)(encObj, buf);
 
// Write out buffer.
	fwrite(buf, 1, bufSiz, fout);
}


static void completeEncWrite2Blks
(	FILE *fout, 
	void (*encBuf)(void *encObj, unsigned char *buf),
	void *encObj,
	unsigned char *buf1,
	unsigned char *buf2,
	unsigned short *nBufsOut,
	int bufSiz,
	int inBufSiz,
	int nBytes
)
{	int totBytes, nBytes1, nBytes2, nBytes2diff;
 
// Calulate useful quantities.
	totBytes = inBufSiz + nBytes;
	nBytes1 = totBytes / 2;
	nBytes2 = totBytes - nBytes1;
	nBytes2diff = nBytes2 - nBytes;
 
// Move bytes in buf2 nBytes2Diff bytes along.
	memmove(buf2+nBytes2diff, buf2, nBytes);
 
// Transfer the last nBytes2Diff of buf1 to the vacated space in buf2.
	memcpy(buf2, buf1+nBytes1, nBytes2diff);
 
// Pocess each buf.
	completeEncWriteBlk(fout, encBuf, encObj, buf1, nBufsOut, bufSiz, nBytes1);
	completeEncWriteBlk(fout, encBuf, encObj, buf2, nBufsOut, bufSiz, nBytes2);
}
	

void eob_close(eob_t *eob)
{
// Freee the zip, and send the remaining chars out.
	if (eob->gz)
		gzipPush_free(eob->gz);

// Encrypt the buffers and write them out.
	if (!eob->isVerbatim)
	{	if (eob->bufNum == 0)
		{ // There was only one buffer written into, i.e. buf[0].
			if (eob->bufPos > 0)
				completeEncWriteBlk(eob->fout, eob->encBuf, eob->encObj, eob->bufs[0],
												&eob->nBufsOut, eob->bufSiz, eob->bufPos);
			else
			{ 	/* No bytes were written ever. */
			}
		}
		else if (eob->bufPos == 0)
		{ // The first buffer is full, but no bytes were written into the next buffer.
			completeEncWriteBlk(eob->fout, eob->encBuf, eob->encObj, eob->bufs[0],
											&eob->nBufsOut, eob->bufSiz, eob->inBufSiz);
		}
		else
		{	// The early buffer is full and there is one partially full buffer.
			// Share the existing bytes between the two buffers and write out.
			completeEncWrite2Blks(eob->fout, eob->encBuf, eob->encObj,
								  eob->bufs[0], eob->bufs[1], &eob->nBufsOut,
								  eob->bufSiz, eob->inBufSiz, eob->bufPos);
		}
	}
	 
// We are done.  Free resources.
	if (eob->encObj != NULL)
		(eob->enc->destroy)(eob->encObj);
	if (eob->buf != NULL)
		free(eob->buf);
	free(eob);
}


static void eob_putChar(void *outStream, unsigned char ch)
{	int pos;
	unsigned char *tmp;
	eob_t *eob = (eob_t*) outStream;
 
// Buffer it.
	eob->bufs[eob->bufNum][eob->bufPos++] = ch;
 
// If the buffer is not full yet, then we are finished.
	if (eob->bufPos < eob->inBufSiz)
		return;
 
// If this is the first buf, then switch to the next buf and we are done.
	if (eob->bufNum == 0)
	{	eob->bufPos = 0;
		eob->bufNum++;
		return;
	}
	
// Now we have two full buffers, and cant store any more.  Must write out.
// Process the earlier buf.
	completeEncWriteBlk(eob->fout, eob->encBuf, eob->encObj, eob->bufs[0],
									&eob->nBufsOut, eob->bufSiz, eob->bufPos);
 
// Switch the buffers so that we write into the space now vacated.
	tmp = eob->bufs[0];
	eob->bufs[0] = eob->bufs[1];
	eob->bufs[1] = tmp;
	eob->bufPos = 0;
 
// We are done.
	return;
}


inline int eob_putc(eob_t *eob, int ch)
{	if (eob->isVerbatim)
	{ // There is no encryption.
		return putc(ch, eob->fout);
	}
	else
	{ // Compress and encrypt.
		gzipPush_outCh((void*)eob->gz, ch);
		return ch;
	}
}


eob_t *eob_open(FILE *fout, int whichEnc, char *key)
{	eob_t *eob;
 
// Allocate eob object.
	eob = malloc(sizeof(eob_t));
	if (eob == NULL)
	{	strcpy(enc_errMsg, "Memory allocation failure 2");
		enc_errStatus = 1;
		return NULL;
	}

// Set resources to NULL.
	eob->enc = NULL;
	eob->buf = NULL;
	eob->encObj = NULL;
	eob->gz = NULL;
	
// Is the encryption valid?
	if (whichEnc<-1 || whichEnc>=mEncs || (whichEnc>=0 && encs[whichEnc].encNum!=whichEnc))
	{	strcpy(enc_errMsg, "Invalid Encryption 1");
		enc_errStatus = 1;
		free(eob);
		return NULL;
	}
 
	if (whichEnc == -1)
	{ // No encryption at all.
		eob->isVerbatim = TRUE;
	}
	else
	{
	// Set up for encryption.
		eob->isVerbatim = FALSE;
 
	// Things for this encryption.
		eob->whichEnc = whichEnc;
		eob->enc = &encs[whichEnc];
		eob->encObj = (eob->enc->create)(key);
		eob->encBuf = eob->enc->encBuf;
		eob->bufSiz = eob->enc->bufSize;
 
	// Buf size without data size information.
		if (eob->bufSiz <= 256) 
			eob->inBufSiz = eob->bufSiz - 1;
		else
			eob->inBufSiz = eob->bufSiz - 2;
		eob->buf = malloc(2*eob->bufSiz);
 
	// Allocate block for buffers.
		if (eob->buf == NULL)
		{	strcpy(enc_errMsg, "Memory allocation failure 3");
			enc_errStatus = 1;
			free(eob);
			return NULL;
		}
 
	// Set up buffers.
		eob->nBufsOut = 0;
		eob->bufs[0] = eob->buf;
		eob->bufs[1] = eob->buf + eob->bufSiz;
		eob->bufNum = 0;
		eob->bufPos = 0;
 
	// Write intro to file.
		fprintf(fout, "%s%c%d%c", nanoCryptStr, '\0', whichEnc, '\0');

	// Set up compression.
		eob->gz = gzipPush_new(eob_putChar, (void*)eob, 0);
	}
 
	eob->fout = fout;
	enc_errStatus = 0;
	return eob;
}


int eob_fwrite(eob_t *eob, char *data, int data_len)
{	int i;
	for (i=0; i<data_len; i++)
		eob_putc(eob, data[i]);
	return data_len;
}


