
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <zlib.h>

#include "gzip.h"


#define gzip_BufSize 16384

#define CKCK {fprintf(CK_fp, "Got to %d %s\n", __LINE__, __FILE__); ck_np++<9000?0:exit(0); }  /**/
extern int ck_np;
extern FILE *CK_fp;  /**/
 
typedef unsigned char gzUchar;

static void error_handle(const char *errmsg)
// Default memory allocator for cstr.
{   fprintf(stderr, "Error in cstr: %s\n", errmsg);
    exit(1);
}


static void *alloc(size_t size)
// Private memory allocator for cstr.
{   void *mem = malloc(size);
    if (mem == NULL)
        error_handle("Memory allocation failure");
    return mem;
}


// =============================== str =================================


typedef struct str_t
{   gzUchar *chars; // The characters of the string.  Extra space for null char.
    int mchars;  // The size of the array 'chars'.
    int nchars;  // The length of the string.
} str_t;


static str_t *str_new(const char *str)
{   str_t *this;
 
// Get the structure. 
    this = (str_t*)alloc(sizeof(str_t));
 
// Get length of the string. 
    if (str == NULL)
        this->nchars = 0;
    else
        this->nchars = strlen(str);
 
// Allocate chars for the string. 
	if (this->nchars == 0)
	{	this->mchars = 0;
		this->chars = NULL;
	}
	else
    {	this->mchars = this->nchars + 8;
		this->mchars |= 7;  // Round up - Use that which would be wasted. 
		this->chars = (gzUchar*)alloc((this->mchars + 1) * sizeof(gzUchar));  // '\0'. 
	}
 
// Copy the string. 
    if (this->nchars > 0)
        memcpy(this->chars, str, this->nchars);
 
// Home with the bacon. 
    return this;
}


static void str_free(str_t *this)
{   if (this->chars)
		free(this->chars);
    free(this);
}


static void str_append_ch(str_t *this, gzUchar ch)
{   int new_len;
 
// Get new length of the string. 
    new_len = this->nchars + 1;
 
// Allocate chars for the string. 
    if (this->mchars < new_len)
    {   this->mchars = this->mchars * 2 + 10;
        this->mchars |= 7;  // Round up - Use that which would be wasted. 
        this->chars = (gzUchar*)realloc(this->chars, (this->mchars+1)*sizeof(gzUchar));
        if (this->chars == NULL)
        {   error_handle("Memory allocation failure");
            exit(1);
        }
    }
 
// Copy the string. 
    {   this->chars[this->nchars] = ch;
        this->nchars = new_len;
    }
}


static void str_truncate(str_t *this, int len)
{	if (len>=0 && len<this->nchars)
		this->nchars = len;
}


static const gzUchar *str_charr(const str_t *this)
{   if (this->chars == NULL)
	{	return "";
	}
	else
	{	this->chars[this->nchars] = '\0';
		return this->chars;
	}
}


int str_length(const str_t *this)
{   return this->nchars;
}


// =============================== gzipPush =================================

typedef struct gzipPush_t
{	int bufSize;
	int isDec;  // Is decompressing.
	void *outStream; // Output stream.
	void (*putChar)(void *outStream, gzUchar ch);
	gzUchar *inBuf;  int nInBuf;
    gzUchar *outBuf;  int nOutBuf;
    z_stream zip;
} gzipPush_t;


gzipPush_t *gzipPush_new
(	void (*putChar)(void *outStream, gzUchar ch)
,	void *outStream
,	int isDecompress
)
{	int ret;
	gzipPush_t *gz = (gzipPush_t *)alloc(sizeof(gzipPush_t));
	gz->outStream = outStream;
	gz->putChar = putChar;
	gz->isDec = isDecompress;
 
// Initialize the buffers.
	gz->bufSize = gzip_BufSize;
	gz->inBuf = (gzUchar*) alloc(gz->bufSize);
	gz->nInBuf = 0;
	gz->outBuf = (gzUchar*) alloc(gz->bufSize);
	gz->nOutBuf = 0;
 
// Initialise zlib deflation.
    gz->zip.zalloc = Z_NULL;
    gz->zip.zfree = Z_NULL;
    gz->zip.opaque = Z_NULL;
	if (isDecompress)
	{	ret = inflateInit(&gz->zip);
		assert(ret == Z_OK);
	}
	else
	{	ret = deflateInit(&gz->zip, Z_DEFAULT_COMPRESSION);
		assert(ret == Z_OK);
	}
 
// Bye.
	return gz;
}


static void gzipPush_do(gzipPush_t *gz, int isFlush)
{   int ret;
    unsigned int have;
 
// Set up the input.
	gz->zip.avail_in = gz->nInBuf;
	gz->zip.next_in = gz->inBuf;
 
// Process.
	do
	{  
	// Initialise the output.
		gz->zip.avail_out = gz->bufSize;
		gz->zip.next_out = gz->outBuf;
 
	// One round of processing.
		if (gz->isDec)
		{
            // fprintf(stderr, "avail_in=%d\n", gz->zip.avail_in);
			ret = inflate(&gz->zip, Z_NO_FLUSH);
			assert(ret!=Z_NEED_DICT);
			assert(ret!=Z_DATA_ERROR);
			assert(ret!=Z_MEM_ERROR);
		}
		else
		{
            // fprintf(stderr, "avail_in=%d  isFlush=%d\n", gz->zip.avail_in, isFlush);
			ret = deflate(&gz->zip, isFlush);
			assert(ret != Z_STREAM_ERROR);
		}
 
	// Send the results out.
		have = gz->bufSize - gz->zip.avail_out;
		for (int i=0; i<have; i++)
		{	gz->putChar(gz->outStream, gz->outBuf[i]);
		}
 
	// Do it, for as long as we write a full block.
	// If the block was not full, then that was the final one.
	} while (gz->zip.avail_out == 0);
	assert(gz->zip.avail_in == 0);
}


void gzipPush_free(gzipPush_t *gz)
{
// Process the remainder.
	if (gz->nInBuf > 0)
		gzipPush_do(gz, Z_FINISH);
 
// Free resources.
	deflateEnd(&gz->zip);
	free(gz->inBuf);
	free(gz->outBuf);
	free(gz);
}


// Push one char out for compression or decompression.
void gzipPush_outCh(gzipPush_t *gz, gzUchar ch)
{	gz->inBuf[gz->nInBuf++] = ch;
	if (gz->nInBuf == gz->bufSize)
	{	gzipPush_do(gz, Z_NO_FLUSH);
		gz->nInBuf = 0;
	}
}


// ---------  Testing --------- .
// 
// void putChar(void *outStream, gzUchar ch)
// {	putc(ch, (FILE*)outStream);
// }
// 
// 
// int main(int argc, gzUchar **argv)
// {	int ch, isDecompress;
// 
// // Are we compressing or decompressing?
// 	isDecompress = 0;
// 	if (argc==2 && strcmp(argv[1],"-d") == 0)
// 		isDecompress = 1;
// 
// // Initialise gzip.
// 	gzipPush_t *gz = gzipPush_new((void *)stdout, putChar, gzipPush_SuggestedBufSize, isDecompress);
// 
// // Xfer the data.
// 	while ((ch=getchar()) != EOF)
// 		gzipPush_outCh(gz, (gzUchar)ch);
// 
// // Free resources.
// 	gzipPush_free(gz);
// }


//==============================  gzipPull =================================


typedef struct gzipPull_t
{	str_t *str;
	int strLen;
	int strNdx;
	const gzUchar *buf;
	gzipPush_t *gz;
	int (*getCh)(void *inStream);
	void *inStream;
} gzipPull_t;


static void gzipPull_putc(void *outStream, gzUchar ch)
{	str_append_ch((str_t*)outStream, ch);
}


gzipPull_t *gzipPull_new(int (*getCh)(void *inStream), void *inStream, int isDecompress)
{	gzipPull_t *gp = (gzipPull_t *)malloc(sizeof(gzipPull_t));
	gp->str = str_new(NULL);
	gp->strLen = 0;
	gp->strNdx = 0;
	gp->buf = NULL;
	gp->gz = gzipPush_new(gzipPull_putc, (void*)gp->str, isDecompress);
	gp->getCh = getCh;
	gp->inStream = inStream;
	return gp;
}


void gzipPull_free(gzipPull_t *gp)
{	if (gp->gz)
		gzipPush_free(gp->gz);
	str_free(gp->str);
	free(gp);
}


int gzipPull_getc(gzipPull_t *gp)
{	
// If there is a char in the buffer, then just get it.
	if (gp->strNdx < gp->strLen)
	{
		return(gp->buf[gp->strNdx++]);
	}
 
// If there is nothing in the buffer, and no more input, then we are finished.
	gzipPush_t *gz = gp->gz;
	if (gz == NULL)
		return -1;
 
// Clean out the buffer.
	str_t *str = gp->str;
	str_truncate(str, 0);
 
// Push input chars to gzip until some output appears.
	void *inStream = gp->inStream;
	int ch = gp->getCh(inStream);
	while (ch!=-1 && str_length(str)==0)
	{	gzipPush_outCh(gz, ch); 
		ch = gp->getCh(inStream);
	}
 
// If there is no more input, then close the zipper.
	if (ch == -1)
	{	gzipPush_free(gz);
		gp->gz = NULL;
	}
	else
	{	gzipPush_outCh(gz, ch); 
	}
 
// Set up the buffer.
	gp->strLen = str_length(str);
// fprintf(stderr, "slen=%d\n", gp->strLen);
	gp->strNdx = 0;
	gp->buf = str_charr(str);
 
// Return a character if there is one.
	if (gp->strNdx < gp->strLen)
		return(gp->buf[gp->strNdx++]);
 
// There is nothing in the buffer.
	return -1;
}


// ---------  Testing --------- .
// 
// int getChar(void *inStream)
// {	return getc((FILE*)inStream);
// }
// 
// 
// int main(int argc, char **argv)
// {	int ch, isDecompress;
//  
// // Are we compressing or decompressing?
// 	isDecompress = 0;
// 	if (argc==2 && strcmp(argv[1],"-d") == 0)
// 		isDecompress = 1;
//  
// // Initialise gzip.
// 	gzipPull_t *gp = gzipPull_new(getChar, (void*)stdin, isDecompress);
//  
// // Xfer the data.
// 	while ((ch=gzipPull_getc(gp)) != -1)
// 		putchar(ch);
//  
// // Free resources.
// 	gzipPull_free(gp);
// }



//==============================  testZip.sh  =================================
// 
// sum1=$(cs4 -d $1)
// ./a.out < $1 | ./a.out -d > $HOME/temp1
// sum2=$(cs4 -d $HOME/temp1)
// if [[ $sum1 == $sum2 ]]
// then
// 	echo -n OK
// else
// 	echo -n BAD
// fi
// echo "" $1


