#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "rota.h"

// Sizes.
#define ROT_MASK 31		/* Must be 2^n-1 for same n. */

// Initial spoke positions.
#define ROTI0 0
#define ROTI1 1
#define ROTI2 3
#define ROTI3 7
#define ROTI4 15

// Values to add.
#define rotA0 151
#define rotA1 149
#define rotA2 139
#define rotA3 137
#define rotA4 131

// Values to multiply.
#define rotM0 307
#define rotM1 317
#define rotM2 337
#define rotM3 353
#define rotM4 367

struct rota_s
{	int i0, i1, i2, i3, i4;
	unsigned short sd[ROT_SIZE];
	unsigned short save_sd[ROT_SIZE];
	int save_i[5];
};


static inline unsigned short screwshort(unsigned short val)
{	return 
	  ((val&0x00ff) << 8)
	+ ((val&0x0f00) >> 4)
	+ ((val&0x3000) >> 10)
	+ ((val&0x4000) >> 13)
	+ ((val&0x8000) >> 15) ;
}


static void rota_rotor(rota_t* rota, unsigned char chin)
// This is the core of rota.
{	unsigned short s;
 
// Blend the data.
	rota->sd[rota->i0] = s = ((rota->sd[rota->i0] + rotA0 + chin) * rotM0);
	rota->sd[rota->i1] = ((rota->sd[rota->i1] + rotA1 + screwshort(s)) * rotM1);
	rota->sd[rota->i2] = ((rota->sd[rota->i2] + rotA2) * rotM2) ^ s;
	rota->sd[rota->i3] = (rota->sd[rota->i3] + rotA3 + s) * rotM3;
	rota->sd[rota->i4] = ((rota->sd[rota->i4] + rotA4) ^ s) * rotM4;
 
// Rotate the spokes.
	rota->i0 = (rota->i0+1) & ROT_MASK;
	rota->i1 = (rota->i1+1) & ROT_MASK;
	rota->i2 = (rota->i2+1) & ROT_MASK;
	rota->i3 = (rota->i3+1) & ROT_MASK;
	rota->i4 = (rota->i4+1) & ROT_MASK;
}


unsigned char rota_mask(rota_t *rota, unsigned char chin)
{	unsigned char chout = chin +
			(rota->sd[rota->i0] + rota->sd[rota->i1] + rota->sd[rota->i2]);
	rota_rotor(rota, chin);
	return chout;
}


unsigned char rota_unmask(rota_t *rota, unsigned char chout)
{	unsigned char chin = chout + 1024 -
		(unsigned char)(rota->sd[rota->i0] + rota->sd[rota->i1] + rota->sd[rota->i2]);
	rota_rotor(rota, chin);
	return chin;
}


rota_t *rota_create(unsigned char *key)
{	int is;		/* iseed */
	int ik;		/* ikey */
	int i;
	int keylen;		
	rota_t *rota;
 
/* Allocate the structure. */
	rota = malloc(sizeof(rota_t));
	if (rota == NULL)
		return NULL;
 
/* Initialise to clean */
	rota->i0=ROTI0;
	rota->i1=ROTI1;
	rota->i2=ROTI2;
	rota->i3=ROTI3;
	rota->i4=ROTI4;
	for (is=0; is<ROT_SIZE; is++)
		rota->sd[is] = 0;
 
/* Incorporate key */
	keylen = strlen((char*)key);		
	ik = 0;
	for (is=ROT_SIZE*10; is--;)
	{	if (ik > keylen)
			ik = 0;
		rota_rotor(rota, key[ik++]);
	}

/* Keep a copy of the rotor. */
	for (i=0; i<ROT_SIZE; i++)
		rota->save_sd[i] = rota->sd[i];
	rota->save_i[0] = rota->i0;
	rota->save_i[1] = rota->i1;
	rota->save_i[2] = rota->i2;
	rota->save_i[3] = rota->i3;
	rota->save_i[4] = rota->i4;
 
/* Home with the bacon. */
	return rota;
}


void rota_destroy(rota_t *rota)
{	free(rota);
}


void rota_reset(rota_t *rota)
/* Restore rota from copy. */
{	int i;

	for (i=0; i<ROT_SIZE; i++)
		rota->sd[i] = rota->save_sd[i];
	rota->i0 = rota->save_i[0];
	rota->i1 = rota->save_i[1];
	rota->i2 = rota->save_i[2];
	rota->i3 = rota->save_i[3];
	rota->i4 = rota->save_i[4];
}


unsigned char* rota_digest(rota_t *rota)
{	return (unsigned char*) rota->save_sd;
}


// // ------- Testing for Rota -----------------
// void main(int argc, char **argv)
// {	
// 	char text[] = "The quick brown fox jumped over the lazy dogs tail.";
// 	char *key = "bollokybol";
// 	int len = strlen(text);
// 	int i, ch;
// 	rota_t *rota;
//  
// 	rota = rota_create(key);
// 	for (i=0; i<len; i++)
// 	{	text[i] = rota_mask(rota, text[i]);
// 	}
// 	rota_destroy(rota);
//  
// 	rota = rota_create(key);
// 	for (i=0; i<len; i++)
// 	{	text[i] = rota_unmask(rota, text[i]);
// 	}
// 	rota_destroy(rota);
//  
// 	printf("%s\n", text);
// }



// ------- Testing for ScrewShort -----------------
// void binout(FILE *fp, unsigned long aa)
// /* This function will print the binary representation of 'aa'
//  * without leading zeroes to the stream 'fp'.
//  */
// {	unsigned long msk = 1, a;
// 	a = aa;
// 	while(a >>= 1)
// 		msk <<= 1;
// 	while(msk)
// 	{	if(aa & msk)
// 			putc('1', fp);
// 		else
// 			putc('0', fp);
// 		msk >>= 1;
// 	}
// }
// void main(int argc, char **argv)
// {	
// 	binout(stdout, screwshort(0x8000)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x4000)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x2000)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x1000)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x0800)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x0400)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x0200)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x0100)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x0080)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x0040)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x0020)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x0010)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x0008)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x0004)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x0002)); putc('\n',stdout);
// 	binout(stdout, screwshort(0x0001)); putc('\n',stdout);
// }
// 1
// 10
// 100
// 1000
// 10000
// 100000
// 1000000
// 10000000
// 100000000
// 1000000000
// 10000000000
// 100000000000
// 1000000000000
// 10000000000000
// 100000000000000
// 1000000000000000

